/*
 * DBBrwColMod.prg
 * Control Columnas del TDBBrowse moderno
 *
 * Copyright 2020 Ignacio Ortiz de Ziga
 * Copyright 2020 https://Xailer.com
 * All rights reserved
 *
 */

#include "Xailer.ch"

//------------------------------------------------------------------------------

CLASS XDbBrwColMod FROM TBrwColMod

PUBLISHED:
   ACCESS oDataset      IS oParent:oDataset

   PROPERTY oDataField  AS TDataField ;
                           WRITE METHOD SetFieldObject ;
                           EDITOR PE_DataField

PUBLIC:
   PROPERTY cDataField INIT "" READONLY

RESERVED:
   PROPERTY nDataCol   INIT 0

   METHOD SetFieldObject( oField, lRefresh )
   METHOD Value( nRecord )
   METHOD SetValue( xValue )
   METHOD Edit()
   METHOD DefaultValue()
   METHOD ResetDataField()
   METHOD IsDataField() INLINE ( ValType( ::oDataField ) == "O" )
   METHOD Free()

END CLASS

//------------------------------------------------------------------------------

METHOD SetFieldObject( xField, lRefresh ) CLASS XDbBrwColMod

   LOCAL oField

   DEFAULT lRefresh TO !Empty( ::oParent:Handle )

   IF Valtype( xField ) == "C"
      ::cDataField := Upper( xField )
      IF ::oDataSet != Nil
         oField := ::oDataset:oFieldByName( xField )
      ENDIF
   ELSEIF ValType( xField ) == "O"
      ::cDataField := Upper( xField:cName )
      oField := xField
   ENDIF

   IF oField != Nil
      ::FoDataField := oField
      ::nDataCol    := oField:nIndex
      IF lRefresh
         ::oParent:Refresh()
      ENDIF
      RETURN .T.
   ENDIF

	::FoDataField := xField

RETURN .F.

//------------------------------------------------------------------------------

METHOD Value( nRecord ) CLASS XDbBrwColMod

   LOCAL Value
   LOCAL nOld

   IF !Empty( nRecord )
      nOld := ::oDataset:Recno()
      ::oDataset:Goto( nRecord, .F. )
   ENDIF

   IF ::IsDataField()
      Value := ::oDataField:Value
   ENDIF

   IF !Empty( nRecord )
      ::oDataset:Goto( nOld, .F. )
   ENDIF

RETURN Value

//------------------------------------------------------------------------------

METHOD SetValue( xValue ) CLASS XDbBrwColMod

   IF ::IsDataField()
      IF ValType( xValue ) == ::oDataField:cType
         ::oDataField:Value := xValue
      ENDIF
   ENDIF

RETURN xValue
//------------------------------------------------------------------------------

METHOD Edit() CLASS xDbBrwColMod
   ::Super:Edit()

   IF ::oDataset == NIL .OR. !::IsDataField() .Or. ::nEditType = beNONE
      ::oGridEdit:lReadOnly := .T.
      ::oGridEdit:lTabStop:=.F.
   ENDIF

RETURN NIL

//------------------------------------------------------------------------------

METHOD DefaultValue() CLASS XDbBrwColMod

   LOCAL Value
   LOCAL cType

   IF ::uDefaultValue != Nil
      Value := ::uDefaultValue
   ELSE
      IF ::IsDataField()
         cType := ::oDataField:cType
      ELSE
         cType := "C"
      ENDIF
      SWITCH cType
      CASE "C"
      CASE "M"
      CASE "B"
         Value := ""
         EXIT
      CASE "N"
         Value := 0
         EXIT
      CASE "L"
         Value := .F.
         EXIT
      CASE "D"
         Value := Ctod( "" )
         EXIT
      CASE "T"
         Value := hb_Ctot( "" )
         EXIT
      OTHERWISE
         Value := ""
      END SWITCH
   ENDIF

RETURN Value

//------------------------------------------------------------------------------

METHOD ResetDataField() CLASS XDbBrwColMod

   IF ::IsDataField()
      ::FoDataField := ::FoDataField:cName
   ENDIF

RETURN ::FoDataField

//------------------------------------------------------------------------------

METHOD Free() CLASS XDBBrwColMod

   ::oDataField := Nil

RETURN ::Super:Free()
